/*      cscope - interactive C symbol or text cross-reference
 *
 *      command functions
 */

#include <ctype.h>      /* isdigit*/
#include "global.h"

BOOL    *change;                /* change this line */
BOOL    changing = NO;          /* changing text */
char    newpat[PATLEN + 1];     /* new pattern */
char    pattern[PATLEN + 1];    /* symbol or text pattern */

BOOL    changestring(), docommand(), search();
void    cantfindmsg(), display(), editall(), editref(), help(), mark(), setfield();
/* command loop */

void
command()
{
        int     c;

        /* set the initial cursor position */
        setfield();
        
        /* see if there is a mouse interface */
        mouseinit();

        /* initialize display parameters */
        dispinit();

        /* change to raw input mode but keep output cooked */
        tty_push(RAWINPUT);     /* waits for output to drain */

        /* pause before clearing the screen if there have been error messages */
        if (errorsfound == YES) {
                errorsfound = NO;
                askforchar();
        }
        /* until the quit command is entered */
        display();              /* will dump waiting output */
        while ((c = getchar()) != EOF && c != ctrl('D') && c != ctrl('Z')) {
                
                /* execute the commmand, updating the display if necessary */
                if (docommand(c) == YES) {
                        display();
                }
        }
}

/* execute the command */

static  BOOL
docommand(c)
register int    c;
{
        char    msg[MSGLEN + 1];        /* message */
        register int    i;
        register FILE   *file;

        switch (c) {
                
        case ctrl('R'): /* rebuild the cross reference */
                if (isuptodate == YES) {
                        putmsg("The -d option prevents rebuilding the cross-reference");
                        atfield();
                        return(NO);
                }
                else {  /* restore the terminal state */
                        clearscreen();  /* will dump pending output */
                        tty_push(COOKED);
                        
                        /* remove the included files from the list */
                        while (nsrcfiles > osrcfiles) {
                                free(srcfiles[--nsrcfiles]);
                                free(srcnames[nsrcfiles]);
                        }
                        /* rebuild the cross-reference */
                        close(symrefs);
                        build();
                        if (errorsfound == YES) {
                                errorsfound = NO;
                                askforchar();
                        }
                        tty_pop();
                        symrefs = open(reffile, O_RDONLY);
                        
                        /* revert to the initial display */
                        if (refsfound != NULL) {
                                fclose(refsfound);
                                refsfound = NULL;
                        }
                }
                break;
        case ctrl('X'): /* mouse selection */
                switch (getchar()) {
                case ctrl(']'):         /* sweep */
                        i = 4;
                        break;
                case ctrl('_'):         /* click */
                        i = 2;
                        break;
                default:                /* unknown control sequence */
                        return(NO);
                }
                while (i-- > 0) {
                        getchar();      /* skip a character */
                }
                c = getchar() - 31;     /* get the line number */

                /* if this is a line selection */
                if (c < FLDLINE - 1) {

                        /* find the selected line */
                        /* note: the selection is forced into range */
                        for (i = disprefs - 1; i > 0; --i) {
                                if (c >= displine[i]) {
                                        break;
                                }
                        }
                        /* display it in the file with the editor */
                        editref(i);
                }
                else {  /* this is an input field selection */
                        field = c - FLDLINE;
                        
                        /* force it into range */
                        if (field < 0) {
                                field = 0;
                        }
                        else if (field >= FIELDS) {
                                field = FIELDS - 1;
                        }
                        goto newfield;
                }
                break;
        case '\t':      /* go to next input field */
        case '\n':
        case '\r':
        case ctrl('N'):
                field = (field + 1) % FIELDS;
                goto newfield;
        case ctrl('P'): /* go to previous input field */
                field = (field + (FIELDS - 1)) % FIELDS;
        newfield:
                setfield();
                atfield();
                return(NO);
        case ' ':       /* display next page */
        case '+':
        case ctrl('V'):
                /* don't redisplay if there are no lines */
                if (totallines == 0) {
                        return(NO);
                }
                /* note: seekline() is not used to move to the next 
                 * page because display() leaves the file pointer at
                 * the next page to optimize paging through refs
                 */
                break;
        case '-':       /* display previous page */
                /* don't redisplay if there are no lines */
                if (totallines == 0) {
                        return(NO);
                }
                /* go back two pages because already at next page */
                nextline -= 2 * mdisprefs;
                if (nextline < 1) {
                        nextline = 1;
                }
                seekline(nextline);
                break;
        case '>':       /* append the lines to a file */
                if (totallines == 0) {
                        putmsg("There are no lines to print");
                }
                else {  /* get the file name */
                        at(PRLINE, 1);
                        printf("Append to file: ");
                        if (getline(newpat, sizeof(newpat), NO) > 0) {

                                /* open the file */
                                if ((file = fopen(newpat, "a")) == NULL) {
                                        sprintf(msg, "Cannot open file: %s", newpat);
                                        putmsg(msg);
                                }
                                else {  /* append the lines */
                                        seekline(1);
                                        while ((c = getc(refsfound)) != EOF) {
                                                putc(c, file);
                                        }
                                        seekline(topline);
                                        fclose(file);
                                }
                        }
                        /* clear the file name prompt */
                        at(PRLINE, 1);
                        clearline();
                }
                /* return to the previous field */
                atfield();
                return(NO);
        case '\f':      /* redraw page */
                seekline(topline);
                break;
        case '!':       /* shell escape */
                at(PRLINE, 1);
                menucleanup();
                execute(shell, shell, 0);
                menuinit();     /* menu may be changed by command */
                seekline(topline);
                break;
        case '?':       /* help */
                help();
                break;
        case ctrl('E'): /* edit all lines */
                editall();
                break;
        case '.':       /* repeat last pattern */
        case ctrl('Y'):
                if (pattern[0] != '\0') {
                        printf("%s", pattern);
                        goto repeat;
                }
                break;
        case '\\':      /* next character is not a command */
        case ctrl('Q'):
                putchar('\\');  /* display the quote character */

                /* get a character from the terminal */
                if ((c = getchar()) == EOF) {
                        return(NO);     /* quit */
                }
                printf("\b \b");        /* erase the quote character */
                goto ispat;
        default:
                /* edit a selected line */
                if (isdigit(c) && c != '0' && mouse == NO) {
                        editref(c - '1');
                }
                /* if this is the start of a pattern */
                else if (isprint(c)) {
        ispat:          ungetc(c, stdin);
                        if (getline(newpat, sizeof(newpat), YES) > 0) {
                                strcpy(pattern, newpat);
        repeat:
                                if (field == CHANGE) {
                                        
                                        /* prompt for the new text */
                                        at(PRLINE, 1);
                                        printf("To: ");
                                        getline(newpat, sizeof(newpat), NO);
                                }
                                /* search for the pattern */
                                if (search() == YES) {
                                        switch (field) {
                                        case EDIT:
                                                topline = 1;
                                                editref(0);
                                                break;
                                        case CHANGE:
                                                return(changestring());
                                        }
                                }
                        }
                        else {  /* no pattern--the input was erased */
                                return(NO);
                        }
                }
                else {  /* control character */
                        return(NO);
                }
        }
        return(YES);
}
/* change one text string to another */

static  BOOL
changestring()
{
        char    newfile[PATHLEN + 1];   /* new file name */
        char    oldfile[PATHLEN + 1];   /* old file name */
        char    linenum[NUMLEN + 1];    /* file line number */
        char    msg[MSGLEN + 1];        /* message */
        FILE    *script;                /* shell script file */
        BOOL    anymarked = NO;         /* any line marked */
        int     c, i;
        char    *s;

        /* open the temporary file */
        if ((script = fopen(temp2, "w")) == NULL) {
                putmsg("Cannot open temporary file");
                return(NO);
        }
        /* create the line change indicators */
        change = (BOOL *) mycalloc(totallines, sizeof(BOOL));
        changing = YES;
        menuinit();

        /* until the quit command is entered */
        for (;;) {
                /* display the current page of lines */
                display();
        same:
                /* get a character from the terminal */
                if ((c = getchar()) == EOF || c == ctrl('D') || c == ctrl('Z')) {
                        break;  /* change lines */
                }
                /* see if the input character is a command */
                switch (c) {
                case ' ':               /* display next page */
                case '+':
                case ctrl('V'):
                case '-':               /* display previous page */
                case '\f':              /* redraw page */
                case '!':               /* shell escape */
                case '?':               /* help */
                        docommand(c);
                        break;
                case ESC:               /* don't change lines */
                case ctrl('G'):
                        goto nochange;
                case '*':               /* mark/unmark all displayed lines */
                        for (i = 0; topline + i < nextline; ++i) {
                                mark(i);
                        }
                        atchange();
                        goto same;
                case 'a':               /* mark all lines */
                        for (i = 0; i < totallines; ++i) {
                                change[i] = YES;
                        }
                        /* show that all have been marked */
                        seekline(totallines);
                        break;
                case ctrl('X'):         /* mouse selection */
                        switch (getchar()) {
                        case ctrl(']'):         /* sweep */
                                i = 4;
                                break;
                        case ctrl('_'):         /* click */
                                i = 2;
                                break;
                        default:                /* unknown control sequence */
                                goto same;
                        }
                        while (i-- > 0) {
                                getchar();      /* skip a character */
                        }
                        c = getchar() - 31;     /* get the line number */
        
                        /* find the selected line */
                        /* note: the selection is forced into range */
                        for (i = disprefs - 1; i > 0; --i) {
                                if (c >= displine[i]) {
                                        break;
                                }
                        }
                        mark(i);
                        goto prompt;
                default:
                        /* if a line was selected */
                        if (isdigit(c) && c != '0' && mouse == NO) {
                                mark(c - '1');
                        }
                prompt:
                        atchange();
                        goto same;
                }
        }
        /* for each line containing the old text */
        fprintf(script, "ed - <<\\!\n");
        oldfile[0] = '\0';
        seekline(1);
        for (i = 0; fscanf(refsfound, "%s%*s%s%*[^\n]", newfile, linenum) == 2;
            ++i) {
                /* see if the line is to be changed */
                if (change[i] == YES) {
                        anymarked = YES;
                
                        /* if this is a new file */
                        if (strcmp(newfile, oldfile) != 0) {
                                
                                /* make sure it can be changed */
                                if (access(newfile, WRITE) != 0) {
                                        sprintf(msg, "Cannot write to file %s", newfile);
                                        putmsg(msg);
                                        anymarked = NO;
                                        break;
                                }
                                /* if there was an old file */
                                if (oldfile[0] != '\0') {
                                        fprintf(script, "w\n"); /* save it */
                                }
                                /* edit the new file */
                                strcpy(oldfile, newfile);
                                fprintf(script, "e %s\n", oldfile);
                        }
                        /* output substitute command */
                        fprintf(script, "%ss/", linenum);       /* change */
                        for (s = pattern; *s != '\0'; ++s) {    /* old text */
                                if (strchr("/\\[.^*$", *s) != NULL) {
                                        putc('\\', script);
                                }
                                putc(*s, script);
                        }
                        putc('/', script);                      /* to */
                        for (s = newpat; *s != '\0'; ++s) {     /* new text */
                                if (strchr("/\\&", *s) != NULL) {
                                        putc('\\', script);
                                }
                                putc(*s, script);
                        }
                        fprintf(script, "/gp\n");               /* and print */
                }
        }
        fprintf(script, "w\nq\n!\n");   /* write and quit */
        fclose(script);

        /* if any line was marked */
        if (anymarked == YES) {
                
                /* edit the files */
                clearscreen();
                printf("Changed lines:\n\n");
                execute("sh", "sh", temp2, 0);
                putchar('\n');
                askforchar();
                seekline(1);
        }
        else {
nochange:
                at(PRLINE, 1);
                clearline();
                atfield();
        }
        changing = NO;
        menuinit();
        free((char *) change);
        return(anymarked);
}

/* mark/unmark this displayed line to be changed */

static  void
mark(i)
int     i;
{
        int     j;
        
        j = i + topline - 1;
        if (j < totallines) {
                at(displine[i], 2);
                if (change[j] == NO) {
                        change[j] = YES;
                        putchar('>');
                }
                else {
                        change[j] = NO;
                        putchar(' ');
                }
        }
}
