/*      cscope - interactive C symbol cross-reference
 *
 *      build cross-reference file
 */

#include <ctype.h>      /* isalpha */
#include "global.h"

BOOL    errorsfound = NO;       /* prompt before clearing messages */
BOOL    lex;                    /* lex program indicator */
int     symbols;                /* number of symbols */

static  char    *filename;      /* file name for warning messages */

void
crossref(srcfile, output)
char    *srcfile;
FILE    *output;
{
        register int    i, j;
        register int    length;         /* symbol length */
        int     token;                  /* current token */
        BOOL    too_many = NO;          /* too many symbols in a line */
        struct  {                       /* symbol data */
                int     type;           /* type */
                int     first;          /* index of first character in text */
                int     last;           /* index of last+1 character in text */
                int     length;         /* symbol length */
        } symbol[SYMBOLMAX];
        void    resetscanner();

        /* open the source file */
        if ((yyin = fopen(srcfile, "r")) == NULL) {
                fprintf(stderr, "cscope: cannot open file %s\n", srcfile);
                errorsfound = YES;
                return;
        }
        /* see if this is a lex file */
        lex = NO;
        if (srcfile[strlen(srcfile) - 1] == 'l') {
                lex = YES;
        }
        filename = srcfile;     /* save the file name for warning messages */
        putfilename(output, srcfile);   /* output file name */
        fprintf(output, "\n\t%c<global>\n", FCNDEF);

        /* read the source file */
        resetscanner(); /* minimize carry-over of scanner problem */
        symbols = 0;
        for (;;) {
                /* get the next token */
                switch (token = yylex()) {
                default:
                        /* see if the symbol is already in the list */
                        length = last - first;
                        for (i = 0; i < symbols; ++i) {
                                if (length == symbol[i].length) {
                                        if (strncmp(yytext + first,
                                            yytext + symbol[i].first, length) == 0) {
                                                break;
                                        }
                                }
                        }
                        if (i == symbols) {     /* if not already in list */

                                /* save the symbol if there is room */
                                if (symbols < SYMBOLMAX) {
                                        symbol[symbols].type = token;
                                        symbol[symbols].first = first;
                                        symbol[symbols].last = last;
                                        symbol[symbols].length = length;
                                        ++symbols;
                                }
                                else {  /* too many symbols */
                                        too_many = YES;
                                }
                        }
                        break;
                case NEWLINE:   /* end of line */
                
                        /* adjust the line number for the \n; restore later */
                        --yylineno;
                        
                case LEX_EOF:   /* end of file; last line may not have \n */
                        
                        /* check for too many symbols */
                        if (too_many == YES) {
                                too_many = NO;
                                warning("too many symbols in this line");
                        }
                        /* output cross-reference symbols */
                        for (i = 0; i < symbols; ++i) {
                                
                                /* output any mark */
                                if (symbol[i].type != IDENT) {
                                        putc('\t', output);
                                        putc(symbol[i].type, output);
                                }
                                /* output the symbol */
                                for (j = symbol[i].first; j < symbol[i].last; ++j) {
                                        putc(yytext[j], output);
                                }
                                putc('\n', output);
                        }
                        /* output the source line number and text */
                        if (symbols > 0) {
                                fprintf(output, " %d %s", yylineno, yytext);
                        }
                        symbols = 0;
                        
                        /* check for end of file */
                        if (token == LEX_EOF) {
                                fclose(yyin);   /* close the source file */
                                yylineno = 1;   /* reset the line count */

                                /* output the leading tab expected by the next call */
                                putc('\t', output);
                                return;
                        }
                        /* restore the line number */
                        ++yylineno;
                        break;
                }
        }
}

/* output the file name */

void
putfilename(output, srcfile)
char    *srcfile;
FILE    *output;
{
        /* check for disk out of space */
        if (fprintf(output, "%c%s", NEWFILE, srcfile) < 0) {
                perror("not built because of write error");
                fclose(output);
                unlink(newreffile);
                exit(1);
        }
}

/* print a warning message with the file name and line number */

void
warning(text)
char *text;
{
        fprintf(stderr, "cscope: \"%s\", line %d: warning: %s\n", filename, 
                yylineno, text);
}
