/*      cscope - interactive C symbol cross-reference
 *
 *      directory searching functions
 */

#include "global.h"
#include "ndir.h"

#define SRCINC  20      /* source file list size increment */

char    currentdir[PATHLEN + 1];        /* current directory */
char    **incdirs;      /* #include directories */
char    **srcdirs;      /* source directories */
char    **srcfiles;     /* source files */
char    **srcnames;     /* source file names */
int     nincdirs;       /* number of #include directories */
int     nsrcdirs;       /* number of source directories */
int     nsrcfiles;      /* number of source files */

static  int     msrcfiles = SRCINC;     /* maximum number of source files */

BOOL    infilelist();
char    *compath(), *findsrcfile();
void    addincdir(), addsrcfile();
/* make the source directory list */

void
makesrcdirlist()
{
        char    *suffix;        /* current dir path after first view path dir */
        char    *vpath;         /* view path */
        register char   *s;
        register int    i, n;
        
        /* get the current directory name */
        if (getwd(currentdir) == NULL) {
                fprintf(stderr, "cscope: warning: cannot get current directory name\n");
                strcpy(currentdir, "<unknown>");
        }
        /* get the view path */
        if ((vpath = getenv("VPATH")) != NULL) {
                
                /* match its first node to the first part of the current dir */
                for (i = 0; vpath[i] == currentdir[i] && vpath[i] != '\0'; ++i) {
                        ;
                }
                if ((vpath[i] != ':' && vpath[i] != '\0') ||
                    (currentdir[i] != '/' && currentdir[i] != '\0')) {
                        fprintf(stderr, "cscope: ignoring VPATH because its first node doesn't contain this directory\n");
                        goto noviewpath;
                }
                suffix = &currentdir[i];
        }
        else {
noviewpath:
                vpath = ".";
                suffix = "";
        }
        /* count the nodes in the view path */
        nsrcdirs = 1;
        for (i = 0; vpath[i] != '\0'; ++i) {
                if (vpath[i] == ':') {
                        ++nsrcdirs;
                }
        }
        /* create the source directory list */
        srcdirs = (char **) mymalloc(nsrcdirs * sizeof(char *));
        
        /* split the view path into nodes */
        s = stralloc(vpath);    /* don't change VPATH in environment */
        for (i = 0; *s != '\0'; ++i) {
                srcdirs[i] = s;
                while (*s != '\0' && *++s != ':') {
                        if (*s == '\n') {
                                *s = '\0';
                        }
                }
                if (*s != '\0') {
                        *s++ = '\0';
                }
        }
        /* convert the view path nodes to source directories */
        srcdirs[0] = ".";       /* first src dir is always current dir */
        n = strlen(suffix) + 1;
        for (i = 1; i < nsrcdirs; ++i) {
                s = mymalloc(strlen(srcdirs[i]) + n);
                strcpy(s, srcdirs[i]);
                strcat(s, suffix);
                srcdirs[i] = s;
        }
}
/* make the source file list */

void
makefilelist(argc, argv, namefile)
int     argc;
char    *argv[];
char    *namefile;
{
        DIR     *dirfile;               /* directory file descriptor */
        struct  direct  *entry;         /* directory entry pointer */
        FILE    *names;                 /* name file pointer */
        register char   *file;
        char    path[PATHLEN + 1];
        char    *s;
        int     i;

        /* create the source file list */
        if (argc > 0) {
                msrcfiles += argc;
        }
        srcfiles = (char **) mymalloc(msrcfiles * sizeof(char *));
        srcnames = (char **) mymalloc(msrcfiles * sizeof(char *));

        /* if there is a file of source file names */
        if (namefile != NULL) {
                if ((names = fopen(namefile, "r")) == NULL) {
                        fprintf(stderr, "cscope: cannot open file %s\n", namefile);
                        exit(1);
                }
                /* get the names in the file */
                while (fscanf(names, "%s", path) == 1) {
                        if ((s = findsrcfile(path)) != NULL) {
                                addsrcfile(path, s);
                        }
                        else {
                                fprintf(stderr, "cscope: cannot find file %s\n",
                                    path);
                                errorsfound = YES;
                        }
                }
                fclose(names);
        }
        /* else if there are source file arguments */
        else if (argc > 0) {
                
                /* put them in an list that can be expanded */
                for (i = 0; i < argc; ++i) {
                        file = argv[i];
                        if (infilelist(file) == NO) {
                                if ((s = findsrcfile(file)) != NULL) {
                                        addsrcfile(file, s);
                                }
                                else {
                                        fprintf(stderr, "cscope: cannot find file %s\n",
                                            file);
                                        errorsfound = YES;
                                }
                        }
                }
        }
        else {  /* make a list of all the source files in the directories */
                for (i = 0; i < nsrcdirs; ++i) {
        
                        /* open the directory */
                        if ((dirfile = opendir(srcdirs[i])) == NULL) {
                                fprintf(stderr, "cscope: cannot open directory %s\n",
                                    srcdirs[i]);
                                errorsfound = YES;
                        }
                        else {  /* read each entry in the directory */
                                while ((entry = readdir(dirfile)) != NULL) {
                                        
                                        /* if it is a source file not already found */
                                        file = entry->d_name;
                                        if (entry->d_ino != 0 && 
                                            gmatch(file, "*.[chlyCGHL]") &&
                                            infilelist(file) == NO) {

                                                /* add it to the list */
                                                sprintf(path, "%s/%s", srcdirs[i], file);
                                                addsrcfile(file, path);
                                        }
                                }
                                closedir(dirfile);
                        }
                }
        }
}

/* add a #include directory to the list for each possible source directory */

void
includedir(dir)
char    *dir;
{
        char    path[PATHLEN + 1];
        register int    i;

        /* see if it is a full path name */
        if (dir[0] == '/') {
                addincdir(dir);
        }
        else {  /* compute its path from all the source directories */
                for (i = 0; i < nsrcdirs; ++i) {
                        sprintf(path, "%s/%s", srcdirs[i], dir);
                        addincdir(path);
                }
        }
}

/* add a #include directory to the list */

static  void
addincdir(dir)
char    *dir;
{
        if (incdirs == NULL) {
                incdirs = (char **) mymalloc(sizeof(char *));
        }
        else {
                incdirs = (char **) myrealloc(incdirs, (nincdirs + 1) * sizeof(char *));
        }
        incdirs[nincdirs++] = stralloc(compath(dir));
}

/* add an include file to the source file list */

void
incfile(file, type)
char    *file;
char    type;
{
        char    path[PATHLEN + 1];
        char    *s;
        register int    i;
        
        /* see if the file is already in the source file list */
        if (infilelist(file) == YES) {
                return;
        }
        /* look in current directory if it was #include "file" */
        if (type == '"' && (s = findsrcfile(file)) != NULL) {
                addsrcfile(file, s);
        }
        else {
                /* search for the file in the #include directory list */
                for (i = 0; i < nincdirs; ++i) {
                        sprintf(path, "%s/%s", incdirs[i], file);
                        if (access(path, READ) == 0) {
                                addsrcfile(file, path);
                                break;
                        }
                }
        }
}

/* see if the file is already in the list */

static  BOOL
infilelist(file)
char    *file;
{
        register int    i;
        
        for (i = 0; i < nsrcfiles; ++i) {
                if (strcmp(file, srcnames[i]) == 0) {
                        return(YES);
                }
        }
        return(NO);
}

/* search for the file in the source directory list */

static  char *
findsrcfile(file)
char    *file;
{
        static  char    path[PATHLEN + 1];
        register int    i;
        
        if (file[0] == '/') {
                return(file);
        }
        for (i = 0; i < nsrcdirs; ++i) {
                sprintf(path, "%s/%s", srcdirs[i], file);
                if (access(path, READ) == 0) {
                        return(path);
                }
        }
        return(NULL);
}

/* add a source file to the list */

static  void
addsrcfile(name, path)
char    *name;
char    *path;
{
        /* make sure there is room for the file  */
        if (nsrcfiles == msrcfiles) {
                msrcfiles += SRCINC;
                srcfiles = (char **) myrealloc(srcfiles, msrcfiles * sizeof(char *));
                srcnames = (char **) myrealloc(srcnames, msrcfiles * sizeof(char *));
        }
        /* add the file to the list */
        srcfiles[nsrcfiles] = stralloc(compath(path));
        srcnames[nsrcfiles++] = stralloc(name);
}
