/*      cscope - interactive C symbol cross-reference
 *
 *      display functions
 */

#include <ctype.h>      /* toupper */
#include <setjmp.h>     /* jmp_buf */
#include <signal.h>     /* SIGINT */
#include "global.h"

#ifdef  BSD             /* toupper does not test for an upper- */
#undef  toupper         /* case character in BSD */
#define toupper(c)      (isupper(c) ? (c) : (c)-'a'+'A')        
#endif

#ifdef  CTRACE          /* prevent cursor movement when tracing */
#define at(a, b)        putchar('\n')
#define clearscreen()
#endif

int     *displine;              /* screen line of displayed reference */
int     disprefs;               /* displayed references */
int     field;                  /* input field */
int     mdisprefs;              /* maximum displayed references */
int     nextline;               /* next line to be shown */
int     topline = 1;            /* top line of page */
int     totallines;             /* total reference lines */

static  int     fldline;                /* input field line */
static  int     fldcolumn;              /* input field column */
static  int     filelen;                /* file name display field length */
static  int     fcnlen;                 /* function name display field length */
static  jmp_buf env;                    /* setjmp/longjmp buffer */
static  int     lastdispline;           /* last displayed reference line */
static  char    lastmsg[MSGLEN + 1];    /* last message displayed */
static  int     numlen;                 /* line number display field length */
static  int     searchcount;            /* count of files searched */
static  char    helpstring[] = "Press the ? key for help";
static  char    selprompt[] = 
        "Select lines to change (press the ? key for help): ";

#if defined(u370) | defined(BSD)        /* compiler bug work-around */
#define VOID    int
#else
#define VOID    void
#endif

VOID    calling(), calledby(), finddef(), findsymbol(), findstring(), findfile();

static  struct  {               /* text of input fields */
        char    *text1;
        char    *text2;
        VOID    (*findfcn)();
} fields[] = {
        "List references to this", "C symbol",          findsymbol,
        "Edit", "this function or #define",             finddef,
        "List", "functions called by this function",    calledby,
        "List", "functions calling this function",      calling,
        "List lines containing this", "text string",    findstring,
        "Change this", "text string",                   findstring,
        "List", "file names containing this text string", findfile,
};
        
/* initialize display parameters */

void
dispinit()
{
        /* calculate the maximum displayed reference lines */
        lastdispline = FLDLINE - 4;
        mdisprefs = lastdispline - REFLINE + 1;
        if (mouse == NO && mdisprefs > 9) {
                mdisprefs = 9;
        }
        /* allocate the displayed line array */
        displine = (int *) mymalloc(mdisprefs * sizeof(int));
}
/* display a page of the references */

void
display()
{
        char    file[PATHLEN + 1];      /* file name */
        char    function[PATLEN + 1];   /* function name */
        char    linenum[NUMLEN + 1];    /* line number */
        int     screenline;             /* screen line number */
        int     width;                  /* source line display width */
        register int    i, j;
        register char   *s;

        startbuffering();
        clearscreen();

        /* see if this is the initial display */
        if (refsfound == NULL) {
                printf("Cscope version %s%s", FILEVERSION, FIXVERSION);
                at(1, CO - (int) sizeof(helpstring) + 1);
                printf(helpstring);
        }
        /* if no references were found */
        else if (totallines == 0) {
                
                /* redisplay the last message */
                puts(lastmsg);
        }
        else {  /* display the pattern */
                if (changing == YES) {
                        printf("Change \"%s\" to \"%s\"", pattern, newpat);
                }
                else {
                        printf("%c%s: %s", toupper(fields[field].text2[0]),
                                fields[field].text2 + 1, pattern);
                }
                /* display the column headings */
                at(3, 3);
                printf("%-*s ", filelen, "File");
                if (field < STRING) {
                        printf("%-*s ", fcnlen, "Function");
                }
                if (field < FILENAME) {
                        printf("Line");
                }
                printf("\n\r");

                /* if at end of file go back to beginning */
                if (nextline > totallines) {
                        seekline(1);
                }
                /* calculate the source text column */
                width = CO - filelen - fcnlen - numlen - 5;
                if (field >= STRING) {
                        width += fcnlen + 1;
                }
                /* until the max references have been displayed or 
                   there is no more room */
                topline = nextline;
                for (disprefs = 0, screenline = REFLINE;
                    disprefs < mdisprefs && screenline <= lastdispline;
                    ++disprefs, ++screenline) {
                        
                        /* read the reference line */
                        if (fscanf(refsfound, "%s%s%s %[^\n]", file, function, 
                            linenum, yytext) < 4) {
                                break;
                        }
                        ++nextline;
                        displine[disprefs] = screenline;
                        
                        /* if no mouse, display the selection number */
                        if (mouse == YES) {
                                putchar(' ');
                        }
                        else {
                                printf("%d", disprefs + 1);
                        }
                        /* display any change mark */
                        if (changing == YES && 
                            change[topline + disprefs - 1] == YES) {
                                putchar('>');
                        }
                        else {
                                putchar(' ');
                        }
                        /* display the file name */
                        if (field == FILENAME) {
                                s = file;
                        }
                        else {
                                s = basename(file);
                        }
                        printf("%-*s ", filelen, s);

                        /* display the function name */
                        if (field < STRING) {
                                printf("%-*s ", fcnlen, function);
                        }
                        if (field == FILENAME) {
                                printf("\n\r"); /* go to next line */
                                continue;
                        }
                        /* display the line number */
                        printf("%*s ", numlen, linenum);

                        /* there may be tabs in grep output */
                        while ((s = strchr(yytext, '\t')) != NULL) {
                                *s = ' ';
                        }
                        /* display the source line */
                        s = yytext;
                        for (;;) {

                                /* see if the source line will fit */
                                if ((i = strlen(s)) > width) {
                                        
                                        /* find the nearest blank */
                                        for (i = width; s[i] != ' ' && i > 0; --i) {
                                                ;
                                        }
                                        if (i == 0) {
                                                i = width;      /* no blank */
                                        }
                                }
                                /* print up to this point */
                                printf("%.*s", i, s);
                                s += i;
                                
                                /* if line didn't wrap around */
                                if (i < width) {
                                        printf("\n\r"); /* go to next line */
                                }
                                /* skip blanks */
                                while (*s == ' ') {
                                        ++s;
                                }
                                /* see if there is more text */
                                if (*s == '\0') {
                                        break;
                                }
                                /* indent the continued source line */
                                at(++screenline, CO - width + 1);
                        }
                }
                /* position the cursor for the message */
                j = FLDLINE - 1;
                if (screenline < j) {
                        putchar('\n');
                }
                else {
                        at(j, 1);
                }
                /* check for more references */
                i = totallines - nextline + 1;
                if (i > 0) {
                        printf("* %d more lines - press the space bar to display more *", i);
                }
                /* if this is the last page of references */
                else if (topline > 1 && nextline > totallines) {
                        printf("* Press the space bar to display the first lines again *");
                }
        }
        /* display the input fields */
        at(FLDLINE, 1);
        for (i = 0; i < FIELDS; ++i) {
                printf("%s %s:\n\r", fields[i].text1, fields[i].text2);
        }
        /* display any prompt */
        if (changing == YES) {
                at(PRLINE, 1);
                printf(selprompt);
        }
        else {  /* return to the current input field */
                at(fldline, fldcolumn);
        }
        stopbuffering();
}

/* set the cursor position for the field */
void
setfield()
{
        fldline = FLDLINE + field;
        fldcolumn = strlen(fields[field].text1) + strlen(fields[field].text2) + 4;
}

/* move to the current input field */

void
atfield()
{
        at(fldline, fldcolumn);
}

/* move to the changing lines prompt */

void
atchange()
{
        at(PRLINE, (int) sizeof(selprompt));
}
/* search for the symbol or text pattern */

static
jumpback()
{
        longjmp(env, 1);
}

BOOL
search()
{
        char    file[PATHLEN + 1];      /* file name */
        char    function[PATLEN + 1];   /* function name */
        char    linenum[NUMLEN + 1];    /* line number */
        void     (*savesig)();           /* old value of signal */
        register int    c, i;
        
        /* open the references found file for writing */
        if (refsfound == NULL) {
                if ((refsfound = fopen(temp1, "w")) == NULL) {
                        putmsg("Cannot open temporary file");
                        return(NO);
                }
        }
        else {
                freopen(temp1, "w", refsfound);
        }
        /* find the pattern - stop on an interrupt */
        putmsg("Searching");
        searchcount = 0;
        if (setjmp(env) == 0) {
                savesig = signal(SIGINT, jumpback);
                (*fields[field].findfcn)();
        }
        signal(SIGINT, savesig);

        /* rewind the cross-reference file */
        lseek(symrefs, (long) 0, 0);
        
        /* reopen the references found file for reading */
        freopen(temp1, "r", refsfound);
        nextline = 1;
        totallines = 0;
        
        /* see if it is empty */
        if ((c = getc(refsfound)) == EOF) {
                sprintf(lastmsg, "Could not find the %s: %s", 
                        fields[field].text2, pattern);
                return(NO);
        }
        /* put back the - read */
        ungetc(c, refsfound);
        
        /* count the references found and find the length of the file,
           function, and line number display fields */
        filelen = 4;    /* strlen("File") */
        fcnlen = 8;     /* strlen("Function") */
        numlen = 0;
        while (fscanf(refsfound, "%s%s%s", file, function, linenum) == 3) {
                if ((i = strlen(basename(file))) > filelen) {
                        filelen = i;
                }
                if ((i = strlen(function)) > fcnlen) {
                        fcnlen = i;
                }
                if ((i = strlen(linenum)) > numlen) {
                        numlen = i;
                }
                /* skip the line text */
                while ((c = getc(refsfound)) != EOF && c != '\n') {
                        ;
                }
                ++totallines;
        }
        rewind(refsfound);
        return(YES);
}
/* display search progress */

void
progress()
{
        static  long    start;
        long    now;
        char    msg[MSGLEN + 1];
        long    time();

        /* save the start time */
        if (searchcount == 0) {
                start = time((long *) NULL);
        }
        /* display the progress every 3 seconds */
        else if ((now = time((long *) NULL)) - start >= 3) {
                start = now;
                if (searchcount <= nsrcfiles) {
                        sprintf(msg, "%d of %d source files searched", 
                                searchcount, nsrcfiles);
                }
                else {
                        sprintf(msg, "%d included files searched", 
                                searchcount - nsrcfiles);
                }
                putmsg(msg);
        }
        ++searchcount;
}
/* print error message on system call failure */

void
myperror(text) 
char    *text; 
{
        extern  int     errno, sys_nerr;
        extern  char    *sys_errlist[];
        char    msg[MSGLEN + 1];        /* message */
        register char   *s;

        s = "Unknown error";
        if (errno < sys_nerr) {
                s = sys_errlist[errno];
        }
        sprintf(msg, "%s: %s", text, s);
        putmsg(msg);
}

/* putmsg clears the message line and prints the message */

void
putmsg(msg) 
char    *msg; 
{
        startbuffering();
        at(MSGLINE, 1);
        clearline();
        printf("%s", msg);
        stopbuffering();
        strcpy(lastmsg, msg);
}
/* position references found file at specified line */

void
seekline(line) 
int     line; 
{
        register int    c;

        /* verify that there is a references found file */
        if (refsfound == NULL) {
                return;
        }
        /* go to the beginning of the file */
        rewind(refsfound);
        
        /* find the requested line */
        nextline = 1;
        while (nextline < line && (c = getc(refsfound)) != EOF) {
                if (c == '\n') {
                        nextline++;
                }
        }
}
