/*      cscope - interactive C symbol cross-reference
 *
 *      process execution functions
 */

#include <signal.h>
#include <varargs.h>
#include "global.h"

#define MAXARGS	10

static  void     (*savesig)();           /* old value of signal */

/* execute forks and executes a program or shell script, waits for it to
 * finish, and returns its exit code.
 */

/*VARARGS1*/
execute(va_alist)   /* note: "exec" is already defined on u370 */
va_dcl
{
	va_list ap;
        int     exitcode;
        int     p;
	char *file;
	char *args[MAXARGS];
	int argno = 0;

	va_start(ap);
	file = va_arg(ap, char *);
	while ((args[argno++] = va_arg(ap, char*)) != (char *)0);
	va_end(ap);

        /* fork and exec the program or shell script */
        curs_end();
        tty_push(COOKED);       /* waits for output to drain */
        if ((p = myfork()) == 0) {
                myexecvp(file, args);        /* child */
        }
        else {
                exitcode = join(p);     /* parent */
        }
        curs_begin();           /* must be before tty_pop */
        tty_pop();
        return(exitcode);
}

/* myexecvp is a special interface to the execvp system call to
 * close all files except stdin, stdout, and stderr; and
 * modify argv[0] to reference the last component of its path-name.
 */

static
myexecvp(a, argv) 
char    *a, **argv; 
{
        register int    i;
        
        /* close files */
        for (i = 3; i < getdtablesize() && close(i) == 0; ++i) {
                ;
        }
        /* modify argv[0] to reference the last component of its path name */
        argv[0] = basename(argv[0]);

        /* execute the program or shell script */
        execvp(a, argv);        /* returns only on failure */
        perror("\nExec failed");/* display the reason */
        askforchar();           /* wait until the user sees the message */
        exit(1);                /* exit the child */
}

/* myfork acts like fork but also handles the interrupt and quit signals */

static
myfork() 
{
        int     p;              /* process number */

        p = fork();
        
        /* the parent ignores the interrupt and quit signal */
        if (p > 0) {
                savesig = signal(SIGQUIT, SIG_IGN);
        }
        /* so they can bu used to stop the child */
        else if (p == 0) {
                signal(SIGINT, SIG_DFL);
                signal(SIGQUIT, SIG_DFL);
        }
        /* check for fork failure */
        if (p == -1) {
                myperror("Cannot fork");
        }
        return p;
}

/* join is the compliment of fork */

static
join(p) 
int     p; 
{
        int     status;  
        int     w;

        /* wait for the correct child to exit */
        do {
                w = wait(&status);
        } while (p != -1 && w != p);

        /* restore signal handling */
        signal(SIGQUIT, savesig);

        /* return the child's exit code */
        return(status >> 8);
}

/* get a file's base name from its path name */

char *
basename(path)
char    *path;
{
        char    *s, *strrchr();
        
        if ((s = strrchr(path, '/')) != NULL) {
                return(s + 1);
        }
        return(path);
}
