/*      cscope - interactive C symbol cross-reference
 *
 *      global type, data, and function definitions
 */

#include <fcntl.h>      /* O_RDONLY */
#include <stdio.h>      /* standard I/O package */
#include <sys/types.h>  /* ino_t needed for sys/dir.h */
#include <sys/dir.h>    /* DIRSIZ and direct */
#include "constants.h"  /* misc. constants */
#include "tty.h"        /* COOKED and RAWINPUT */
#include "version.h"    /* VERSION */

#ifdef  BSD
#define strchr  index
#define strrchr rindex
#endif
#ifdef  BSD4_2
#define curdir  getwd
#endif

#define ctrl(x) (x & 037)       /* control character macro */

typedef enum    {               /* boolean data type */
        NO,
        YES
} BOOL;

/* main.c global data */
extern  char    *editor, *home, *shell; /* environment variables */
extern  char    *argv0;         /* command name */
extern  BOOL    isuptodate;     /* consider the crossref up-to-date */
extern  char    *newreffile;    /* new cross-reference file name */
extern  int     osrcfiles;      /* original number of source files */
extern  char    *reffile;       /* cross-reference file path name */
extern  FILE    *refsfound;     /* references found file */
extern  int     symrefs;        /* cross-reference file */
extern  char    temp1[];        /* temporary file name */
extern  char    temp2[];        /* temporary file name */

/* command.c global data */
extern  BOOL    *change;        /* change this line */
extern  BOOL    changing;       /* changing text */
extern  char    newpat[];       /* new pattern */
extern  char    pattern[];      /* symbol or text pattern */

/* crossref.c global data */
extern  BOOL    errorsfound;    /* prompt before clearing error messages */
extern  BOOL    lex;            /* lex program indicator */
extern  int     symbols;        /* number of symbols */

/* cursor.c global data */
extern  int     CO;             /* number of columns in a line */
extern  int     LI;             /* number of lines on screen */
extern  char    *term;          /* terminal type */

/* dir.c global data */
extern  char    currentdir[];   /* current directory */
extern  char    **incdirs;      /* #include directories */
extern  char    **srcdirs;      /* source directories */
extern  char    **srcfiles;     /* source files */
extern  char    **srcnames;     /* source file names */
extern  int     nincdirs;       /* number of #include directories */
extern  int     nsrcdirs;       /* number of source directories */
extern  int     nsrcfiles;      /* number of source files */

/* display.c global data */
extern  int     *displine;      /* screen line of displayed reference */
extern  int     disprefs;       /* displayed references */
extern  int     field;          /* input field */
extern  int     mdisprefs;      /* maximum displayed references */
extern  int     nextline;       /* next line to be shown */
extern  int     topline;        /* top line of page */
extern  int     totallines;     /* total reference lines */

/* find.c global data */
extern  char    *blockp;        /* pointer to current character in block */

/* mouse.c global data */
extern  BOOL    mouse;          /* mouse interface */

/* scanner.l global data */
extern  int     first;          /* buffer index for first char of symbol */
extern  int     last;           /* buffer index for last char of symbol */
extern  FILE    *yyin;          /* input file descriptor */
extern  int     yyleng;         /* length of input line */
extern  int     yylineno;       /* number of input line */
extern  char    yytext[];       /* input line text */

/* library functions */
char    *getenv(), *mktemp();
char    *strcat(), *strcpy(), *strncpy(), *strpbrk(), *strchr(), *strrchr();
long    lseek();
void    exit(), free(), perror(), qsort();
#ifdef  BSD
FILE    *popen();       /* not in stdio.h */
#endif

/* cscope functions called from more than one function */
char    *basename();
char    *nextchar(), *readblock(), *scanpast();
void    build();
void    askforchar(), incfile(), putfilename(), putstring(), warning();
void    myperror(), putmsg(), progress(), seekline();
void    mouseinit(), menucleanup(), menuinit();

/* library functions */
char    *mycalloc(), *mymalloc(), *myrealloc(), *stralloc();
void    curs_begin(), curs_end(), curs_init();
void    at(), clearline(), clearscreen(), startbuffering(), stopbuffering();
void    grepinit();
