/*      cscope - interactive C symbol cross-reference
 *
 *      terminal input functions
 */

#include <ctype.h>      /* iscntrl */
#include <errno.h>      /* EINTR */
#include <signal.h>     /* SIGINT */
#include "tty.h"        /* erasechar and killchar */
#include "global.h"

extern  int     errno;  /* system call error number */

/* catch the interrupt signal */
static
catchint()
{
        /* getchar will return EOF with errno set to EINTR */
}

/* get a line from the terminal in non-canonical mode */

getline(s, n, return_on_erase)
register char   s[];
register int    n;
BOOL    return_on_erase;
{
        void     (*savesig)();           /* old value of signal */
        register int    c, i = 0;
        int     j;

        /* until the end of the line is reached */
        /* note: the check for \r is a work-around for u370's 2.0 ioctl */
        savesig = signal(SIGINT, catchint);
        while ((c = getchar()) != '\n' && c != '\r' && i < n) {
                if (c == erasechar) {   /* erase */
                        if (i > 0) {
                                putchar('\b');
                                putchar(' ');
                                putchar('\b');
                                --i;
                        }
                }
                /* kill or interrupt */
                else if (c == killchar || c == EOF && errno == EINTR) {
                        for (j = 0; j < i; ++j) {
                                putchar('\b');
                        }
                        for (j = 0; j < i; ++j) {
                                putchar(' ');
                        }
                        for (j = 0; j < i; ++j) {
                                putchar('\b');
                        }
                        i = 0;
                }
                else if (c == EOF) {    /* end-of-file */
                        break;
                }
                else if (!iscntrl(c)) { /* if not a control character */
                        putchar(c);     /* display it */
                        s[i++] = c;     /* save it */
                }
                fflush(stdout);
                
                /* return on an empty line to allow a command to be entered */
                if (return_on_erase == YES && i == 0) {
                        break;
                }
        }
        s[i] = '\0';
        signal(SIGINT, savesig);
        return(i);
}

/* ask user to enter a character after reading the message */

void
askforchar()
{
        tty_push(RAWINPUT);
        printf("Type any character to continue: ");
        getchar();
        tty_pop();
}
