/*      cscope - interactive C symbol cross-reference
 *
 *      keyword look-up routine for the C symbol scanner
 */
#include "global.h"

static  struct  keystruct {
        char    *text;
        struct  keystruct *next;
} keyword[] = {
        "auto",         NULL,
        "break",        NULL,
        "case",         NULL,
        "char",         NULL,
        "continue",     NULL,
        "default",      NULL,
        "do",           NULL,
        "double",       NULL,
        "else",         NULL,
        "enum",         NULL,
        "extern",       NULL,
        "float",        NULL,
        "for",          NULL,
        "fortran",      NULL,
        "goto",         NULL,
        "if",           NULL,
        "int",          NULL,
        "long",         NULL,
        "register",     NULL,
        "return",       NULL,
        "short",        NULL,
        "sizeof",       NULL,
        "static",       NULL,
        "struct",       NULL,
        "switch",       NULL,
        "typedef",      NULL,
        "union",        NULL,
        "unsigned",     NULL,
        "void",         NULL,
        "while",        NULL,
};
#define KEYWORDS        (sizeof(keyword) / sizeof(struct keystruct))

#define HASHSIZE        KEYWORDS        /* number of keywords */

static  struct  keystruct *hashtab[HASHSIZE]; /* pointer table */

/* put the keywords into the symbol table */
void
init_symtab()
{
        register int    i;
        
        for (i = 0; i < KEYWORDS; ++i) {
                hashlink(&keyword[i]);
        }
}

/* see if this identifier is a keyword */
int
lookup(ident)
register char   *ident;
{
        register struct keystruct *p;
        
        /* look up the identifier in the keyword table */
        for (p = hashtab[hash(ident)]; p != NULL; p = p->next) {
                if (strcmp(ident, p->text) == 0) {
                        return(KEYWORD);
                }
        }
        /* this is an identifier */
        return(IDENT);
}

/* link to hash table */
static  int
hashlink(p)
register struct keystruct *p;
{
        register int    hashval;
        
        hashval = hash(p->text);
        p->next = hashtab[hashval];
        hashtab[hashval] = p;
}

/* form hash value for string */
static  int
hash(s)
register char   *s;
{
        register int    hashval;
        
        for (hashval = 0; *s != '\0'; )
                hashval += *s++;
        return(hashval % HASHSIZE);
}
