/*      cscope - interactive C symbol cross-reference
 *
 *      main functions
 */

#include <signal.h>     /* SIGINT and SIGPIPE */
#include "global.h"     /* includes sys/types.h needed for sys/stat.h */
#include <sys/stat.h>   /* stat */

/* command defaults */
#define EDITOR  "vi"
#define PAGER   "more"
#define SHELL   "sh"

#define TEMPLEN 23      /* temporary file name length */

char    *editor, *home, *shell; /* environment variables */
char    *argv0;                 /* command name */
BOOL    isuptodate = NO;        /* consider the crossref up-to-date */
char    *newreffile;            /* new cross-reference file name */
int     osrcfiles;              /* original number of source files */
char    *reffile = REFFILE;     /* cross-reference file path name */
FILE    *refsfound;             /* references found file */
int     symrefs = -1;           /* cross-reference file */
char    temp1[TEMPLEN + 1];     /* temporary file name */
char    temp2[TEMPLEN + 1];     /* temporary file name */

main(argc, argv)
int     argc;
char    **argv;
{
        char    file[PATHLEN + 1];      /* file name */
        char    *namefile = NULL;       /* file of file names */
        char    *s;
        int     c, i;
        char    *mygetenv();
        void    command(), includedir(), init_symtab(), 
                makesrcdirlist(), makefilelist();
        void    quit();         /* needed for ctrace */
        
        /* save the command name */
        argv0 = argv[0];

        /* make the source directory list */
        makesrcdirlist();
        
        /* set the options */
        while (--argc > 0 && (*++argv)[0] == '-') {
                for (s = argv[0] + 1; *s != '\0'; s++) {
                        switch (*s) {
                        case '-':       /* end of options */
                                --argc;
                                ++argv;
                                goto lastarg;
                        case 'V':       /* print the version number */
                                fprintf(stderr, "cscope: version %s%s\n", 
                                        FILEVERSION, FIXVERSION);
                                exit(0);
                        case 'd':       /* consider crossref up-to-date */
                                isuptodate = YES;
                                break;
                        case 'f':       /* alternate cross-reference file */
                        case 'i':       /* file containing file names */
                        case 'I':       /* #include file directory */
                                c = *s;
                                if (*++s == '\0' && --argc > 0) {
                                        s = *++argv;
                                }
                                switch (c) {
                                case 'f':       /* alternate cross-reference file */
                                        reffile = s;
                                        break;
                                case 'i':       /* file containing file names */
                                        namefile = s;
                                        break;
                                case 'I':       /* #include file directory */
                                        includedir(s);
                                        break;
                                }
                                goto nextarg;
                        default:
                                fprintf(stderr, "cscope: unknown option: -%c\n", *s);
                                fprintf(stderr, "Usage: cscope [-d] [-f cross-reference file] [-i name file] [-I header directory]\n\t[source files]\n");
                                exit(1);
                        }
                }
nextarg:        ;
        }
lastarg:
        /* read the environment */
        editor = mygetenv("EDITOR", EDITOR);
        editor = mygetenv("VIEWER", editor);    /* use viewer if set */
        home = getenv("HOME");
        shell = mygetenv("SHELL", SHELL);

        /* set up the I/O */
        curs_init();
        tty_init();

        /* put the terminal into cursor motion mode */
        /* note: this prevents emacsterm from paging the build messages */
        curs_begin();   /* must be before mouseinit() */

        /* if the cross-reference is to be considered up-to-date */
        if (isuptodate == YES) {
                FILE    *oldrefs;       /* old cross-reference */
                
                if ((oldrefs = fopen(reffile, "r")) == NULL) {
                        fprintf(stderr, "cscope: cannot open file %s\n", reffile);
                        exit(1);
                }
                /* skip the version and current directory */
                fscanf(oldrefs, "cscope %*s %*s");
                
                /* skip the source directory list */
                fscanf(oldrefs, "%d", &nsrcdirs);
                for (i = 0; i < nsrcdirs; ++i) {
                        fscanf(oldrefs, "%*s");
                }
                /* skip the #include directory list */
                fscanf(oldrefs, "%d", &nincdirs);
                for (i = 0; i < nincdirs; ++i) {
                        fscanf(oldrefs, "%*s");
                }
                /* get the number of source files */
                fscanf(oldrefs, "%d", &nsrcfiles);

                /* get the source file list */
                fprintf(stderr, "cscope: creating list of source files\n");
                srcfiles = (char **) mymalloc(nsrcfiles * sizeof(char *));
                for (i = 0; i < nsrcfiles; ++i) {
                        fscanf(oldrefs, "%s", file);
                        srcfiles[i] = stralloc(file);
                }
                fclose(oldrefs);
        } else {
                /* add /usr/include to the directory list */
                includedir("/usr/include");
        
                /* make the source file list */
                makefilelist(argc, argv, namefile);
                osrcfiles = nsrcfiles;  /* save this for rebuilding */
                if (nsrcfiles == 0) {
                        fprintf(stderr, "cscope: no source files found\n");
                        exit(1);
                }
                /* see if you can create the cross-reference in this dir */
                if (reffile[0] != '/' && access(".", WRITE) != 0) {
        
                        /* no, put it in the home directory */
                        s = mymalloc(strlen(reffile) + strlen(home) + 2);
                        sprintf(s, "%s/%s", home, reffile);
                        reffile = s;
                }
                /* initialize the C keyword table */
                init_symtab();
        
                /* create the file name used for a new cross-reference */
                newreffile = mymalloc(strlen(reffile) + strlen(home) + 3);
                if ((s = strrchr(reffile, '/')) != NULL) {
                        ++s;
                        strcpy(newreffile, reffile);
                        sprintf(newreffile + (s - reffile), "n%s", s);
                }
                else {
                        sprintf(newreffile, "n%s", reffile);
                }
                /* build the cross-reference */
                build();
        }
        /* if running in the foreground */
        if (signal(SIGINT, SIG_IGN) != SIG_IGN) {

                /* ignore interrupt signal */
                signal(SIGINT, SIG_IGN);

                /* cleanup on quit signal */
                signal(SIGQUIT, quit);
        }
        /* open the cross-reference file */
        symrefs = open(reffile, O_RDONLY);
        
        /* create the temporary files */
        i = getpid();
        sprintf(temp1, "/usr/tmp/cscope%d.1", i);
        sprintf(temp2, "/usr/tmp/cscope%d.2", i);
        
        /* enter the command loop */
        command();
        
        /* cleanup and exit */
        quit();
}

/* return the environment value or, if not found, the default arg */

char    *
mygetenv(variable, deflt)
char    *variable, *deflt;
{
        char    *value;

        value = getenv(variable);
        if (value == NULL || strlen(value) == 0) {
                return(deflt);
        }
        return(value);
}
/* build the cross-reference */

void
build()
{
        char    *oldfile;               /* file in old cross-reference */
        char    newdir[PATHLEN + 1];    /* directory in new cross-reference */
        char    olddir[PATHLEN + 1];    /* directory in old cross-reference */
        char    oldname[PATHLEN + 1];   /* name in old cross-reference */
        char    oldversion[FILEVERLEN + 1];     /* version in old cross-ref */
        int     oldnum;                 /* number in old cross-ref */
        struct  stat statstruct;        /* file status */
        time_t  reftime;                /* old crossref modification time */
        FILE    *newrefs;               /* new cross-reference */
        FILE    *oldrefs;               /* old cross-reference */
        int     firstfile;              /* first source file in pass */
        int     lastfile;               /* last source file in pass */
        register char   *cp;
        register int    i;
        int     compare();              /* needed for qsort call */
        void    crossref();
        char    *getoldfile();

        /* normalize the current directory relative to the home directory so
           the cross-reference is not rebuilt when the user's login is moved */
        strcpy(newdir, currentdir);
        if (strcmp(currentdir, home) == 0) {
                strcpy(newdir, "$HOME");
        }
        else if (strncmp(currentdir, home, strlen(home)) == 0) {
                sprintf(newdir, "$HOME%s", currentdir + strlen(home));
        }
        /* sort the source file names (needed for rebuilding) */
        qsort(srcfiles, (unsigned) nsrcfiles, sizeof(char *), compare);

        /* if there is an old cross-reference and its version and current
           directory matches */
        if ((oldrefs = fopen(reffile, "r")) != NULL &&
            fscanf(oldrefs, "cscope %s %s", oldversion, olddir) == 2 &&
            strcmp(oldversion, FILEVERSION) == 0 &&
            (strcmp(olddir, currentdir) == 0 || /* remain compatible */
             strcmp(olddir, newdir) == 0)) {
                
                /* get the cross-reference file's modification time */
                fstat(fileno(oldrefs), &statstruct);
                reftime = statstruct.st_mtime;
                
                /* see if the number of source directories is the same */
                if (fscanf(oldrefs, "%d", &oldnum) < 1 ||
                    oldnum != nsrcdirs) {
                        goto outofdate;
                }
                /* see if the source directory list is the same */
                for (i = 0; i < nsrcdirs; ++i) {
                        if (fscanf(oldrefs, "%s", oldname) < 1 ||
                            strcmp(oldname, srcdirs[i]) != 0) {
                                goto outofdate;
                        }
                }
                /* see if the number of #include directories is the same */
                if (fscanf(oldrefs, "%d", &oldnum) < 1 ||
                    oldnum != nincdirs) {
                        goto outofdate;
                }
                /* see if the #include directory list is the same */
                for (i = 0; i < nincdirs; ++i) {
                        if (fscanf(oldrefs, "%s", oldname) < 1 ||
                            strcmp(oldname, incdirs[i]) != 0) {
                                goto outofdate;
                        }
                }
                /* see if the number of source files is the same */
                if (fscanf(oldrefs, "%d", &oldnum) < 1 ||
                    oldnum != nsrcfiles) {
                        goto outofdate;
                }
                /* see if the list of source files is the same and
                   none have been changed */
                for (i = 0; i < nsrcfiles; ++i) {
                        if (fscanf(oldrefs, "%s", oldname) < 1 ||
                            strcmp(oldname, srcfiles[i]) != 0 ||
                            stat(srcfiles[i], &statstruct) != 0 ||
                            statstruct.st_mtime > reftime) {
                                goto outofdate;
                        }
                }
                /* the old cross-reference is up-to-date */
                fclose(oldrefs);
                return;
                
        outofdate:
                /* reopen the old cross-reference file for fast scanning */
                symrefs = open(reffile, O_RDONLY);

                /* get the first file name in the old cross-reference */
                readblock();            /* read the first cross-ref block */
                scanpast('\t');         /* skip the header */
                oldfile = getoldfile();
        }
        else {  /* force cross-referencing of all the source files */
                reftime = 0;
                oldfile = NULL;
        }
        /* open the new cross-reference file */
        if ((newrefs = fopen(newreffile, "w")) == NULL) {
                fprintf(stderr, "cscope: cannot open file %s\n", newreffile);
                exit(1);
        }
        /* output the cscope version and current directory */
        fprintf(newrefs, "cscope %s %s\n", FILEVERSION, newdir);

        /* output the source directory count and names */
        fprintf(newrefs, "%d ", nsrcdirs);
        for (i = 0; i < nsrcdirs; ++i) {
                fprintf(newrefs, "%s\n", srcdirs[i]);
        }
        /* output the #include directory count and names */
        fprintf(newrefs, "%d ", nincdirs);
        for (i = 0; i < nincdirs; ++i) {
                fprintf(newrefs, "%s\n", incdirs[i]);
        }
        /* output the source file count and names */
        fprintf(newrefs, "%d ", nsrcfiles);
        for (i = 0; i < nsrcfiles; ++i) {
                fprintf(newrefs, "%s\n", srcfiles[i]);
        }
        /* output the leading tab expected by crossref() */
        putc('\t', newrefs);

        /* make passes through the source file list until the last level of
           included files is processed */
        fprintf(stderr, "cscope: symbol cross-reference for\n");
        firstfile = 0;
        lastfile = nsrcfiles;
        for (;;) {

                /* get the next source file name */
                for (i = firstfile; i < lastfile; ++i) {
                        fprintf(stderr, "%s ", srcfiles[i]);    /* print file name */
                        
                        /* if the old file has been deleted get the next one */
                        while (oldfile != NULL && strcmp(srcfiles[i], oldfile) > 0) {
                                oldfile = getoldfile();
                        }
                        /* if there isn't an old database or this is a new file */
                        if (oldfile == NULL || strcmp(srcfiles[i], oldfile) < 0) {
                                crossref(srcfiles[i], newrefs); /* crossref it */
                                fprintf(stderr, "built\n");
                        }
                        /* if this file was modified */
                        else if (stat(srcfiles[i], &statstruct) == 0 &&
                            statstruct.st_mtime > reftime) {
                                crossref(srcfiles[i], newrefs); /* crossref it */
                                fprintf(stderr, "built\n");
                                
                                /* skip its old crossref so modifying the last
                                   source file does not cause all included files
                                   to be built.  Unfortunately a new file that is
                                   alphabetically last will cause all included
                                   files to be build, but this is less likely */
                                oldfile = getoldfile();
                        }
                        else {  /* copy its cross-reference */
                                putfilename(newrefs, srcfiles[i]);
                                setmark('\t');
                                cp = blockp;
                                for (;;) {
                                        /* copy up to the next \t */
                                        do {    /* innermost loop optimized to only one test */
                                                while (*cp != '\t') {
                                                        putc(*cp++, newrefs);
                                                }
                                        } while (*++cp == '\0' && (cp = readblock()) != NULL);

                                        putc('\t', newrefs);    /* copy the \t */
                                        
                                        /* get the next character */
                                        if (*(cp + 1) == '\0') {
                                                cp = readblock();
                                        }
                                        /* exit if at the end of this file's xref */
                                        if (cp == NULL || *cp == NEWFILE) {
                                                break;
                                        }
                                        /* look for an #included file */
                                        if (*cp == INCLUDE) {
                                                putc(INCLUDE, newrefs);
                                                blockp = cp;
                                                nextchar();
                                                putstring(oldname);
                                                fputs(oldname, newrefs);
                                                incfile(oldname + 1, *oldname);
                                                setmark('\t');
                                                cp = blockp;
                                        }
                                }
                                blockp = cp;
                                fprintf(stderr, "copied\n");
                                oldfile = getoldfile();
                        }
                }
                /* see if any included files were found */
                if (lastfile == nsrcfiles) {
                        break;
                }
                firstfile = lastfile;
                lastfile = nsrcfiles;
        
                /* sort the included file names */
                qsort(&srcfiles[firstfile], (unsigned) (lastfile - firstfile),
                        sizeof(char *), compare);
        }
        /* add a null file name to the trailing tab */
        fprintf(newrefs, "%c\n", NEWFILE);
        fclose(newrefs);
        
        /* close and remove the old cross-reference file */
        if (symrefs >= 0) {
                close(symrefs);
        }
        if (oldrefs != NULL) {
                fclose(oldrefs);
                unlink(reffile);
        }
        /* replace it with the new cross-reference file */
        link(newreffile, reffile);
        unlink(newreffile);
}

/* get the next file name in the old cross-reference */

static  char *
getoldfile()
{
        static  char    file[PATHLEN + 1];      /* file name in old crossref */

        if (blockp != NULL) {
                do {
                        if (*blockp == NEWFILE) {
                                nextchar();
                                putstring(file);
                                if (file[0] != '\0') {
                                        return(file);
                                }
                        }
                } while (scanpast('\t') != NULL);
        }
        return(NULL);
}
        
/* string comparison function for qsort */

static  int
compare(s1, s2)
char    **s1;
char    **s2;
{
        return(strcmp(*s1, *s2));
}

/* cleanup and exit */

static  void
quit()
{
        void    curs_cleanup(), tty_cleanup();
        
        /* remove any temporary files */
        if (temp1[0] != '\0') {
                unlink(temp1);
                unlink(temp2);
        }
        /* abort any output */
        clearscreen();
        
        /* restore the terminal to its original mode */
        tty_cleanup();
        curs_end();
        menucleanup();
        exit(0);
}
