/*      cscope - interactive C symbol cross-reference
 *
 *      mouse functions
 */

#include "global.h"

#define MENUMAX 10                      /* maximum length of menu */

BOOL    mouse = NO;                     /* mouse interface */

typedef struct {                        /* menu */
        char    *text;
        char    value;
} MENU;

static  MENU    mainmenu[] = {          /* main menu */
        "Next Lines",           ' ',
        "Previous Lines",       '-',
        "Repeat Search",        '.',
        "Edit All Lines",       ctrl('E'),
        "Append to File",       '>',
        "Rebuild Xref",         ctrl('R'),
        "Start Shell",          '!',
        "Redraw Screen",        ctrl('L'),
        "Help",                 '?',
        "Exit",                 ctrl('D'),
        NULL, '\0'
};

static  MENU    changemenu[] = {        /* change mode menu */
        "Mark Displayed Lines", '*',
        "Next Lines",           ' ',
        "Previous Lines",       '-',
        "Mark All Lines",       'a',
        "Change Marked Lines",  ctrl('D'),
        "Do Not Change Lines",  ESC,
        "Start Shell",          '!',
        "Redraw Screen",        ctrl('L'),
        "Help",                 '?',
        NULL, '\0'
};

static  MENU    *loaded;                /* menu loaded */

void    menuload();
/* see if there is a mouse interface */

void
mouseinit()
{
        char    *mouseterm;
        char    *mygetenv();
        
        /* the MOUSE enviroment variable is for 5620 terminal programs that have
           mouse support but the TERM environment variable is the same as a
           terminal without a mouse, such as myx */
        mouseterm = mygetenv("MOUSE", "");
        if (strcmp(term, "emacsterm") == 0 || strcmp(mouseterm, "myx") == 0) {
                mouse = YES;
        }
        menuload(mainmenu);
}

/* initialize the mouse menu */

void
menuinit()
{
        if (changing == YES) {
                menuload(changemenu);
        }
        else {
                menuload(mainmenu);
        }
}

/* download a menu */

static  void
menuload(menu)
MENU    *menu;
{
        register int    i, len;

        if (mouse == YES) {
                tty_push(COOKED);       /* will lose output without this */
                startbuffering();
                if (strcmp(term, "emacsterm") == 0) {

                        /* switch to menu 2 and/or reset the default selection */
                        printf("\033#2");
                        
                        /* load the menu */
                        for (i = 0; i < MENUMAX && menu[i].text != NULL; ++i) {
                                printf("\033M%d@%s@%c@", i, menu[i].text, 
                                        menu[i].value);
                        }
                        /* if the menu is smaller than the max, terminate it */
                        if (i < MENUMAX) {
                                printf("\033M%d@@@", i);
                        }
                }
                else {  /* myx */
                        
                        /* load the menu */
                        menucleanup();
                        printf("\033[6;1X\033[9;1X");
                        for (i = 0; menu[i].text != NULL; ++i) {
                                len = strlen(menu[i].text);
                                printf("\033[%d;%dx%s%c", len, len + 1, 
                                        menu[i].text, menu[i].value);
                        }
                        loaded = menu;
                }
                stopbuffering();
                tty_pop();
        }
}

/* cleanup any menu */

void
menucleanup()
{
        register int    i;

        if (loaded != NULL) {
                tty_push(COOKED);       /* will lose output without this */
                startbuffering();
                printf("\033[6;0X\033[9;0X");
                for (i = 0; loaded[i].text != NULL; ++i) {
                        printf("\033[0;0x");
                }
                stopbuffering();
                tty_pop();
                loaded = NULL;
        }
}
