%{
/*      cscope - interactive C symbol cross-reference
 *
 *      C symbol scanner
 */
#include "global.h"

/* the line counting has been moved from character reading for speed */
#undef  input
#define input()  ((yytchar=yysptr>yysbuf?*--yysptr:getc(yyin))==EOF?0:yytchar)
#undef  unput
#define unput(c) (*yysptr++=yytchar=(c))

int     first;  /* buffer index for first char of symbol */
int     last;   /* buffer index for last char of symbol */

static  int     braces;         /* unmatched left brace count */
static  int     token;          /* token found */
static  BOOL    fcnbody;        /* function body */
static  BOOL    rules;          /* lex/yacc rules */
static  BOOL    ppdefine;       /* preprocessor define statement */

void    comment(), scanfor();
%}
identifier      [a-zA-Z_][a-zA-Z_0-9]*
number          "."?[0-9][.0-9a-fA-FlLxX]*
space           [ \t]
%%
\%\%            {       /* lex/yacc rules deliminter */
                        braces = 0;
                        if (rules == NO) {
                                rules = YES;
                        }
                        else {
                                rules = NO;
                        }
                }
\{              {       /* count unmatched left braces for fcn def detection */
                        ++braces;
                        goto more;
                }
^\}             {       /* attempt to correct erroneous brace count caused by:
                         * 
                         * #ifdef ...
                         *      ... {
                         * #else
                         *      ... {
                         * #endif
                         */
                        braces = 0;
                        fcnbody = NO;
                        goto more;
                }
\}              {       /* check for the end of a function */
                        if (--braces == 0 && fcnbody == YES) {
                                fcnbody = NO;
                        }
                        goto more;
                }
\#{space}*define{space}+{identifier}    {
                                
                        /* preprocessor macro or constant definition */
                        ppdefine = YES;
                        token = DEFINE;
                        first = yyleng - 1;
                        while (yytext[first] != ' ' && yytext[first] != '\t') {
                                --first;
                        }
                        ++first;
                        goto fcn;
                }
{identifier}/[ \t]*\([^"\n)]*\)[ \t\n]*[{/a-zA-Z_]      {
                        
                        /* warning: "if (...)" must not overflow yytext, so a
                           function definition is not allowed to contain \n */
                        
                        /* look for a function definition */
                        /* note: "#define a (b)" is not one */
                        if (braces == 0 && ppdefine == NO && rules == NO) {
                                fcnbody = YES;
                                token = FCNDEF;
                                goto fcn;
                        }
                        goto fcncal;
                }
{identifier}/[ \t]*\(   { /* look for a function call */
                fcncal: if (fcnbody == YES || ppdefine == YES || rules == YES) {
                                token = FCNCALL;
                                goto fcn;
                        }
                        goto ident;
                }
{identifier}    {
                ident:  token = IDENT;
                
                        /* check for a long line */
                fcn:    if (yyleng > LINEMAX) {
                                int     c;
                                
                                /* skip to the end of the line */
                                warning("line too long");
                                while ((c = input()) != LEX_EOF) {
                                        if (c == '\n') {
                                                unput(c);
                                                break;
                                        }
                                }
                        }
                        /* check for a keyword */
                        yymore();
                        if (lookup(yytext + first) == KEYWORD) {
                                first = yyleng;
                        }
                        else {  /* identifier */
                                last = yyleng;
                                return(token);
                        }
                }
\\\n            {       /* preprocessor statement is continued on next line */
                        goto nl;
                }
\n              {       /* end of the line */
                        ppdefine = NO;
                nl:     ++yylineno;
                        first = 0;
                        last = 0;
                        if (symbols > 0) {
                                return(NEWLINE);
                        }
                }
['"]            {       /* character or string constant */
                        scanfor(yytext[yyleng - 1]);
                        goto more;
                }
"/*"            {       /* delete comments to prevent line buffer overflow */
                        yyleng -= 2;    /* delete the leading comment delimiter */
                        comment();      /* delete the rest plus trailing white space */
                        goto more;      /* this comment may be embedded in a statement */
                }
^[ \t\f]+       {       /* don't save leading white space */
                }
\t              {       /* change a tab to a blank */
                        yytext[yyleng - 1] = ' ';
                        goto more;
                }
\#{space}*include{space}*["<][^">]+     { /* #include file */
                        char    *s;
                        
                        s = strpbrk(yytext, "\"<");
                        incfile(s + 1, *s);
                        first = s - yytext;
                        last = yyleng;
                        yymore();
                        return(INCLUDE);
                }
\#{space}*{identifier}  |       /* preprocessor keyword */
{number}        |       /* number */
.               {       /* punctuation and operators */
                more:   first = yyleng;
                        yymore();
                }
%%

/* don't carry-over scanner problems to new file */
void
resetscanner()
{
        first = 0;      /* buffer index for first char of symbol */
        last = 0;       /* buffer index for last char of symbol */
        braces = 0;     /* unmatched left brace count */
        fcnbody = NO;   /* function body */
        rules = NO;     /* lex/yacc rules */
        ppdefine = NO;  /* preprocessor define statement */
}
static  void
comment()
{
        register char   c, last_c;

        /* find the trailing comment delimiter */
        last_c = '\0';
        while ((c = input()) != LEX_EOF) {
                if (c == '/' && last_c == '*') { /* there are fewer '/'s */
                        break;
                }
                else if (c == '\n') {
                        ++yylineno;
                }
                last_c = c;
        }
}
static  void
scanfor(terminator)
register char   terminator;
{
        register char   c;

        /* scan until the terminator is found */
        while ((c = yytext[yyleng++] = input()) != terminator) {
                switch (c) {
                case '\\':      /* escape character */
                        yytext[yyleng++] = input();
                        break;
                case '\t':      /* tab character */
                
                        /* if not a lex program, change to a blank and continue */
                        if (lex == NO) {
                                yytext[yyleng - 1] = ' ';
                                break;
                        }
                case '\n':      /* illegal character */
                        
                        /* assume the terminator is missing, so put
                           this character back */
                        unput(c);
                        yytext[--yyleng] = '\0';
                        return;
                case LEX_EOF:   /* end of file */
                        return;
                }
                /* see if this token will overflow the line buffer */
                /* note: '\\' may cause yyleng to be > LINEMAX */
                if (yyleng >= LINEMAX) {
                        
                        /* truncate the token */
                        while ((c = input()) != LEX_EOF) {
                                if (c == terminator) {
                                        unput(c);
                                        break;
                                }
                                else if (c == '\n') {
                                        ++yylineno;
                                }
                        }
                }
        }
        yytext[yyleng] = '\0';
}
/* dummy routine to avoid need for lex library on non-UNIX systems */
static
yywrap()
{
        return(1);
}
#ifdef BSD
/*      @(#)strpbrk.c   1.2     */
/*LINTLIBRARY*/
/*
 * Return ptr to first occurance of any character from `brkset'
 * in the character string `string'; NULL if none exists.
 */

#define NULL    (char *) 0

char *
strpbrk(string, brkset)
register char *string, *brkset;
{
        register char *p;

        do {
                for(p=brkset; *p != '\0' && *p != *string; ++p)
                        ;
                if(*p != '\0')
                        return(string);
        }
        while(*string++);
        return(NULL);
}
#endif
