/* terminal mode functions */

/* Modes:

        COOKED--the standard mode upon entry.

        RAWINPUT--no echoing, erase or kill processing.
                XON/XOFF, interrupt, and quit signal procesing is done.
*/

#ifdef BSD
#include <sgtty.h>      /* sgttyb */
#define termio  sgttyb
#define TCGETA  TIOCGETP
#define TCSETAW TIOCSETP
#else
#include <sys/ioctl.h>  /* needed for termio.h */
#include <termio.h>     /* termio */
#endif
#include <stdio.h>      /* fflush(stdout) */
#include "tty.h"        /* COOKED and RAWINPUT */

#define STDIN   0       /* standard input */
#define TTYMAX  10      /* maximum tty mode stack size */

/* termcap library variables set by tty_init */
#if defined(SV_R1) | defined(BSD)
extern  short   ospeed; /* terminal output speed */
#endif

char    erasechar, killchar;            /* input control chars */

static  struct  termio  cookedterm, rawinputterm;
static  int     curmode = -1;           /* current tty mode */
static  int     ttysp = 0;              /* tty mode stack pointer */
static  char    tty_stack[TTYMAX];      /* tty mode stack */

void    tty_set();
/* get the inital terminal mode */

void
tty_init() 
{
        /* get initial (cooked) terminal mode */
        ioctl(STDIN, TCGETA, &cookedterm);
        
        /* get the erase and kill chars for use during raw input */
#ifdef BSD
        erasechar = cookedterm.sg_erase;
        killchar = cookedterm.sg_kill;
#else
        erasechar = cookedterm.c_cc[VERASE];
        killchar = cookedterm.c_cc[VKILL];
#endif
        /* save the terminal output speed for the termcap library */
#ifdef BSD
        ospeed = cookedterm.sg_ospeed;
#endif
#ifdef SV_R1
        ospeed = cookedterm.c_cflag & CBAUD;
#endif
        /* set up raw input mode options */
        rawinputterm = cookedterm;
#ifdef BSD
        rawinputterm.sg_flags |= CBREAK;        /* return char as soon a typed */
        rawinputterm.sg_flags &= ~ECHO; /* no echo */
#else
        rawinputterm.c_lflag = ISIG;            /* allow interrupt and quit */
        rawinputterm.c_cc[VMIN] = 1;            /* minimum characters for read */
        rawinputterm.c_cc[VTIME] = 1;           /* minimum seconds/10 for read */
        rawinputterm.c_iflag = IXON | IXANY;    /* XON/XOFF commands and no interrupts */
        rawinputterm.c_oflag = 0;               /* no output processing */
#endif
}

/* save the old mode on the stack and set the terminal to the new mode */

void
tty_push(pmode) 
int     pmode; 
{
        if (ttysp < TTYMAX - 1) {
                ++ttysp;
                tty_set(pmode);
        }
}

/* restore the previous mode from the stack */

void
tty_pop() 
{
        if (ttysp > 0) {
                tty_set(tty_stack[--ttysp]);
        }
}

/* reset the terminal to cooked mode if tty_push has ever been called */

void
tty_cleanup()
{
        if (curmode >= 0) {
                tty_set(COOKED);
        }
}

/* set the terminal to the requested mode if it is not in that mode */

static  void
tty_set(pmode) 
int     pmode; 
{
        tty_stack[ttysp] = pmode;
        if (pmode != curmode) {
                curmode = pmode;
                switch (pmode) {
                case RAWINPUT:
                        fflush(stdout); /* System V line buffers to terminal */
                        ioctl(STDIN, TCSETAW, &rawinputterm);
                        break;
                case COOKED:
                        ioctl(STDIN, TCSETAW, &cookedterm);
                        break;
                }
        }
}
